#include <iostream>
#include <iomanip>
#include <vector>
#include <fstream>
#include "numerov.h"
#include "zeroin.h"
#include "integrate.h"

using namespace std;

class RadialWave{
  int N;
  vector<double> R;        // Radial mesh but reversed R[0]=Rmax and R.last=0
  vector<double> Solution; // Non-normalized solution of Schroedinger equation
  vector<double> rhs;      // Right-hand-site in solving Schroedinger equation
  vector<double> Veff;     // Effective KS potential
  vector<double> Veff0;    // Effective KS potential without centrifugal part
public:
  RadialWave(const vector<double>& Rmesh) : N(Rmesh.size()), Solution(N), R(N), rhs(N), Veff(N), Veff0(N)
  {
    for (int i=0; i<N; i++) R[i] = Rmesh[N-1-i]; // The mesh is reversed
    Solution[0] = R[0]*exp(-R[0]); // Boundary (starting) points of integration by Numerov
    Solution[1] = R[1]*exp(-R[1]); // Boundary (starting) points of integration by Numerov
  }
  // This function-operator is used to find bound states. It is given to root-finding routine
  double operator()(double E)
  {
    double h = R[1]-R[0];
    for (int i=0; i<N-1; i++) rhs[i] = 2*(Veff[i]-E); // The RHS of the SCHR-equation for choosen energy E
    rhs[R.size()-1]=0;                                // This is the zero frequency
    Numerov(rhs, Solution.size()-1, h, Solution);     // Solving the radial SCH-equation
    int last = Solution.size()-1;                     // The last point at zero frequency needs extrapolation
    Solution[last] = Solution[last-1]*(2+h*h*rhs[last-1])-Solution[last-2];
    return Solution[last];                            // Value at zero frequency
  }
  // This function return the density of electrons (up+down per volume) of one (nl) state.
  void Density(vector<double>& rho)
  {
    rho.resize(Solution.size());
    int N = Solution.size();
    for (int i=0; i<Solution.size(); i++) rho[i] = Solution[N-1-i]*Solution[N-1-i]; // The mesh outside this class is reversed!
    double norm = 1./integrate4<double>(rho, R[0]-R[1], rho.size());                // Normalization constant 
    for (int i=1; i<rho.size(); i++) rho[i] = rho[i]*norm/(4*M_PI*sqr(R[N-i-1]));   // rho_{nl}=u^2/(4*Pi*r^2)
    rho[0] = 2*rho[1]-rho[2];                                                       // extrapolation to zero frequency
  }
  // This function sets KS potential without centrifugal part
  void SetVeff0(const vector<double>& Uhartree, const vector<double>& Vxc, int Z)
  { for (int i=0; i<R.size()-1; i++) Veff0[i] = (-Z+Uhartree[N-1-i])/R[i]+Vxc[N-1-i];}
  void AddCentrifugal(int l)
  { for (int i=0; i<R.size()-1; i++) Veff[i] = Veff0[i] + 0.5*l*(l+1)/sqr(R[i]);}
  void SetVeff(const vector<double>& Uhartree, const vector<double>& Vxc, int l, int Z)
  { SetVeff0(Uhartree,Vxc,Z);
    AddCentrifugal(l); }
  double V_KS0(int i){return Veff0[N-1-i];}
};

class ExchangeCorrelation{
//******************************************************************************/
//  Calculates Exchange&Correlation Energy and Potential                       */ 
//  type=0 - due to U.von.Barth and L.Hedin, J.Phys.C5, 1629 (1972)            */
//  type=1 - O.E.Gunnarsson and S.Lundqvist,  Phys.Rev.B                       */
//  type=2 - V.L.Moruzzi, J.F.Janak, and A.R.Williams, Calculated              */
//           Electronic Properties of Metals (New York, Pergamon Press, 1978)  */
//  type=3 - S.H.Vosko, L.Wilk, and M.Nusair, Can.J.Phys.58, 1200 (1980)       */
//  type=4 - Correlation of Perdew and Wang 1991                               */
//******************************************************************************/
  int type;
  double A, C;
  static const double alphax = 0.610887057710857;//(3/(2 Pi))^(2/3)
  static const double Aw = 0.0311;
  static const double Bw = -0.048;
  static const double Cw = 0.002;
  static const double D  = -0.0116;
  static const double gamma  = -0.1423;
  static const double beta1  =  1.0529;
  static const double beta2  =  0.3334;
  static const double Ap  =  0.0621814;
  static const double xp0 = -0.10498;
  static const double bp  =  3.72744;
  static const double cp  =  12.9352;
  static const double Qp  =  6.1519908;
  static const double cp1 =  1.2117833;
  static const double cp2 =  1.1435257;
  static const double cp3 = -0.031167608;
public:
  ExchangeCorrelation(int type_) : type(type_)
  {
    switch(type){
    case 0: C = 0.0504; A = 30; break;
    case 1: C = 0.0666; A = 11.4; break;
    case 2: C = 0.045;  A = 21; break;
    }
  };
  double Vx(double rs){return -alphax/rs;}
  double ExVx(double rs){return 0.25*alphax/rs;}
  double Ex(double rs){return -0.75*alphax/rs;}
  double Vc(double rs)
  {
    if (type<3){
      double x = rs/A;
      return -0.5*C*log(1+1/x);
    }else if(type<4){// type=3 WVN
      double x=sqrt(rs);
      double xpx=x*x+bp*x+cp;
      double atnp=atan(Qp/(2*x+bp));
      double ecp = 0.5*Ap*(log(x*x/xpx)+cp1*atnp-cp3*(log(sqr(x-xp0)/xpx)+cp2*atnp));
      return ecp - Ap/6.*(cp*(x-xp0)-bp*x*xp0)/((x-xp0)*xpx);
    }else{
      if (rs>1) return gamma/(1+beta1*sqrt(rs)+beta2*rs)*(1+7/6.*beta1*sqrt(rs)+beta2*rs)/(1+beta1*sqrt(rs)+beta2*rs);
      else return Aw*log(rs)+Bw-Aw/3.+2/3.*Cw*rs*log(rs)+(2*D-Cw)*rs/3.;
    }
  }
  double EcVc(double rs)
  {
    if (type<3){
      double x = rs/A;
      double epsilon = -0.5*C*((1+x*x*x)*log(1+1/x)+0.5*x-x*x-1/3.);
      return epsilon-Vc(rs);
    } else if (type<4){// type=3 WVN
      double x=sqrt(rs);
      return Ap/6.*(cp*(x-xp0)-bp*x*xp0)/((x-xp0)*(x*x+bp*x+cp));
    }else{
      if (rs>1) return 2*gamma/(1+beta1*sqrt(rs)+beta2*rs)-Vc(rs);
      else return Aw*log(rs)+Bw+Cw*rs*log(rs)+D*rs-Vc(rs);
    }
  }
};

// Calculates numer of principal quantum numbers n to be used for certain Z
int NumberOfPrincipal(int Z)
{
  double v = pow((54*Z+sqrt(-3+2916.*Z*Z))*3.,1/3.);
  int n0 = static_cast<int>((-3+3/v+v)/6.+0.99);
  if (Z>36) n0++;                // Because 5s is filled before 4f
  return n0;
}

// Class for storing the quantum numbers n,l, and Energy together
// It is important to be able to sort those states by the energy
class BState{
public:
  int n, l;
  double E;
  BState(){E=0;}
  BState(int n_, int l_, double E_): n(n_), l(l_), E(E_){};
};
// Class for comparison of above classes. Needed to sort bound states by the energy
class Cmp{
public:
  int operator()(const BState& a, const BState& b){return a.E<b.E;}
};

void SolvePoisson(int Zq, const vector<double>& Rmesh, const vector<double>& rho, vector<double>& Uhartree)
{// Given the input density rho, calculates the Hartree potential
  static vector<double> RHS(Rmesh.size());
  for (int i=0; i<Rmesh.size(); i++) RHS[i] = -4*M_PI*Rmesh[i]*rho[i];
  Uhartree[0]=0;  Uhartree[1]=(Rmesh[1]-Rmesh[0]);// Boundary condition for U_H=V_H/r
  NumerovInhom(RHS, RHS.size(), Rmesh[1]-Rmesh[0], Uhartree); // Solving the 2nd order differential equation
  // adding homogeneous solution to satisfay boundary conditions: U(0)=0, U(infinity)=Z
  int ilast = Uhartree.size()-1;
  double U_last = Uhartree[ilast];
  double alpha = (Zq - U_last)/Rmesh[ilast];
  for (int i=0; i<Rmesh.size(); i++) Uhartree[i] += alpha*Rmesh[i];
}

/////////////////////////////////////////////
////// n=0	n=1	n=2	n=3	.....
//////---------------------------------------
////// l=0	l=0	l=0	l=0	.....
//////		l=1	l=1	l=1	.....
//////			l=2	l=2	.....
//////				l=3
////// E(n+1,l) > E(n,l) therefore we need to keep
////// track of E(n,l).
////// We start with E(n,l=n) > -0.5*Z^2/(l+1)^2-3.
//////////////////////////////////////////////
void FindBoundStates(int n0, int Z, double dEz, RadialWave& wave, vector<BState>& states)
{// Searches for bound state with given n,l. They are stored in vector<BState>
  vector<vector<double> > zeros(n0);            // Two dimensional function for storing bound-state energies [n,l]
  // this 2D container is necessary to keep track of bound states withs smaller n and same l -- the preceeding bound state
  for (int i=0; i<zeros.size(); i++) zeros[i].resize(n0); //2D needs to be resized, unfortunatrly. Better to used derived class
  for (int i=0; i<states.size(); i++) states[i].E=100;// Sets some high energy not to mix with bound states
  int j=0;
  for (int n=0; n<n0; n++){
    for (int l=0; l<=n; l++){
      wave.AddCentrifugal(l);                        // Adds centrifugal part to effective KS potential
      double x = (n<=l) ? -0.5*Z*Z/sqr(l+1)-3. : zeros[n-1][l]+1e-7; //Starts looking for zero, the first one should be guessed!
      double v0 = wave(x), v1=v0;
      while(x<10.){                                 // Looks for zero even at positive frequencies somethimes
	x+=dEz;                                     // Proceeding in small steps to bracket all zeros
	v1 = wave(x);                               // New value of radial function at origin
	if (v0*v1<0) {                              // Changs sign?
	  double zero = zeroin(x-dEz,x,wave,1e-10); // Root-finder locates bound state very precisely
	  zeros[n][l] = zero;                       // Stores into two places, just to remember where is "n-1"-bound state
	  states[j++] = BState(n,l,zero);           // Stores solution
	  clog<<"Found solutin for n="<<n<<" l="<<l<<" at "<<zeros[n][l]<<endl;
	  break;
	}
      }
    }
  }
}

inline double rs(double rho)
{return pow(3/(4*M_PI*rho),1/3.);}

int main(int argc, char *argv[], char *env[])
{
  int Z, N;
  ifstream input("rho.dat");
  string str;
  input>>str;
  input>>Z;
  input>>N;
  vector<double> Rmesh(N), rho(N), Uhartree(N);;
  // read input density
  for (int i=0; i<N; i++) input>>Rmesh[i]>>rho[i];

  // Solve Poisson equation
  SolvePoisson(Z, Rmesh, rho, Uhartree);// Given the input density rho, calculates the Hartree potential


  ///////////////////////////////////////////////////////////////////////////////////////////////////////
  ////////////////////////// Second part of the code ////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////////////////////////////

  // Initializing class for exchange and correlation energy and potential
  ExchangeCorrelation XC(3);// WVN seems to be the best (look http://physics.nist.gov/PhysRefData/DFTdata/Tables/ptable.html)
  vector<double> Vxc(Rmesh.size()); // Exchange and correlation potential
  // Adding exchange-correlation part
  Vxc[0]=0;
  for (int i=1; i<Rmesh.size(); i++){
    double rs_ = rs(rho[i]);
    Vxc[i] = XC.Vx(rs_) + XC.Vc(rs_); // computes exchange correlation potential from the density
  }

  ///////////////////////////////////////////////////////////////////////////////////////////////////////
  ////////////////////////// Third part of the code /////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////////////////////////////
  // Initialization of class for solving radial Schroedinger equation
  RadialWave wave(Rmesh); 
  // Sets hartree+exchange+correlation part of the potential (not centrifugal part)
  wave.SetVeff0(Uhartree, Vxc, Z);
  
  // obtains the upper limit of main quantum number  
  int n0 = NumberOfPrincipal(Z); // From Z, one can calculate which main quantum numbers n need to be incuded
  double dEz = 0.1;              // Step when searching for bound states
  vector<BState> states(static_cast<int>((Z+1)/2)); // States to be sorted for lowest few to build density from
  FindBoundStates(n0, Z, dEz, wave, states);        // Finds all bound states

  
  Cmp cmp; // class for sorting bound state
  sort(states.begin(),states.end(),cmp); // Sort states
  
  return 0;
}
